library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ws2812test is
	port(
		clk       : in  std_logic;
		reset     : in  std_logic;
		ws_out    : out std_logic;
		gamma_req : in  std_logic
	);
end entity ws2812test;

architecture RTL of ws2812test is
	constant LEDCNT_c : integer := 60;
	constant HEARTBEATMAX_c : integer := 700000;
	signal rst         : std_logic;
	signal ws_busy     : std_logic;
	signal ws_strb     : std_logic;
	signal red_gamma   : unsigned(7 downto 0);
	signal green_gamma : unsigned(7 downto 0);
	signal blue_gamma  : unsigned(7 downto 0);

	-- Signals for renderer
	signal render_active : std_logic;
	type transition_t is (RY, YG, GC, CB, BM, MR);
	signal transition_active : transition_t;
	type color_rec_t is record
		red   : unsigned(7 downto 0);
		green : unsigned(7 downto 0);
		blue  : unsigned(7 downto 0);
	end record color_rec_t;
	signal color_active     : color_rec_t;
	signal lednum           : integer range 0 to LEDCNT_c - 1;
	-- Signals for rainbow
	signal heartbeat        : std_logic;
	signal transition_start : transition_t;
	signal color_start      : color_rec_t;
	signal red_ws           : unsigned(7 downto 0);
	signal green_ws         : unsigned(7 downto 0);
	signal blue_ws          : unsigned(7 downto 0);
begin
	rst <= not reset;

	beatgen : process(clk, rst) is
		variable beat_counter : integer range 0 to HEARTBEATMAX_c := 0;
	begin
		if rst = '1' then
			beat_counter := 0;
			heartbeat    <= '0';
		elsif rising_edge(clk) then
			heartbeat <= '0';
			if beat_counter /= HEARTBEATMAX_c then
				beat_counter := beat_counter + 1;
			else
				beat_counter := 0;
				heartbeat    <= '1';
			end if;
		end if;
	end process beatgen;

	gamma_red : entity work.gamma8bit
		port map(
			color_in  => color_active.red,
			color_out => red_gamma
		);
	gamma_green : entity work.gamma8bit
		port map(
			color_in  => color_active.green,
			color_out => green_gamma
		);

	gamma_blue : entity work.gamma8bit
		port map(
			color_in  => color_active.blue,
			color_out => blue_gamma
		);

	red_ws   <= red_gamma when gamma_req = '1' else color_active.red;
	green_ws <= green_gamma when gamma_req = '1' else color_active.green;
	blue_ws  <= blue_gamma when gamma_req = '1' else color_active.blue;

	ws2812bphy_inst : entity work.ws2812bphy
		generic map(
			HIGH1 => 40,
			LOW1  => 23,
			HIGH0 => 20,
			LOW0  => 43
		)
		port map(
			clk    => clk,
			rst    => rst,
			busy   => ws_busy,
			ws_out => ws_out,
			strb   => ws_strb,
			red    => red_ws,
			green  => green_ws,
			blue   => blue_ws
		);

	rainbow : process(clk, rst) is
		procedure increment_col(signal color : inout unsigned(7 downto 0); is_renderer : boolean; next_state : transition_t) is
		begin
			color <= color + 20;
			if color = 220 then
				if is_renderer then
					transition_active <= next_state;
				else
					transition_start <= next_state;
				end if;
			end if;
		end procedure increment_col;

		procedure decrement_col(signal color : inout unsigned(7 downto 0); is_renderer : boolean; next_state : transition_t) is
		begin
			color <= color - 20;
			if color = 20 then
				if is_renderer then
					transition_active <= next_state;
				else
					transition_start <= next_state;
				end if;
			end if;
		end procedure decrement_col;

	begin
		if rst = '1' then
			render_active      <= '1';
			color_active.red   <= to_unsigned(240, color_start.red'length);
			color_active.green <= x"00";
			color_active.blue  <= x"00";

			color_start.red   <= to_unsigned(240, color_start.red'length);
			color_start.green <= x"00";
			color_start.blue  <= x"00";
			transition_active <= RY;
			transition_start  <= RY;
			lednum            <= 0;
		elsif rising_edge(clk) then
			-- renderer
			if render_active = '1' then
				ws_strb <= '0';
				if ws_busy = '0' and ws_strb = '0' then
					ws_strb <= '1';
					case transition_active is
						when RY =>
							increment_col(color_active.green, true, YG);

						when YG =>
							decrement_col(color_active.red, true, GC);
						when GC =>
							increment_col(color_active.blue, true, CB);
						when CB =>
							decrement_col(color_active.green, true, BM);
						when BM =>
							increment_col(color_active.red, true, MR);
						when MR =>
							decrement_col(color_active.blue, true, RY);
					end case;

				end if;                 -- phy ready
				if ws_strb = '1' then
					if lednum < (LEDCNT_c - 1) then
						lednum <= lednum + 1;
					else
						lednum        <= 0;
						render_active <= '0';
					end if;
				end if;
			elsif heartbeat = '1' then
				case transition_start is
					when RY =>
						increment_col(color_start.green, false, YG);

					when YG =>
						decrement_col(color_start.red, false, GC);
					when GC =>
						increment_col(color_start.blue, false, CB);
					when CB =>
						decrement_col(color_start.green, false, BM);
					when BM =>
						increment_col(color_start.red, false, MR);
					when MR =>
						decrement_col(color_start.blue, false, RY);
				end case;
				color_active      <= color_start;
				transition_active <= transition_start;
				render_active     <= '1';

			end if;                     -- rainbow


		end if;                         -- rising edge

	end process rainbow;

end architecture RTL;
